from ._block import Block
from ._blocker import Blocker
from ._lsh import LSHBlocker
from ._sorted_neighborhood import SortedNeighborhoodBlocker
from ._tf_idf import TfIdfBlocker
from ._ground_truth import GroundTruthBlocker

__all__ = [
    "Block",
    "Blocker",
    "LSHBlocker",
    "SortedNeighborhoodBlocker",
    "TfIdfBlocker",
    "GroundTruthBlocker",
]



from dataclasses import dataclass

from matchescu.reference_store.comparison_space import BinaryComparisonSpace
from matchescu.typing import EntityReferenceIdentifier


@dataclass
class BlockingMetrics:
    pair_completeness: float
    pair_quality: float
    reduction_ratio: float


class BinaryComparisonSpaceEvaluator(object):
    def __init__(
        self,
        gt: set[tuple[EntityReferenceIdentifier, EntityReferenceIdentifier]],
        original_size: int,
    ):
        self.__gt = gt
        self.__size = original_size

    def __call__(self, comparison_space: BinaryComparisonSpace) -> BlockingMetrics:
        candidate_ids = set(comparison_space)
        true_positive_pairs = self.__gt.intersection(candidate_ids)
        pc = len(true_positive_pairs) / len(self.__gt) if len(self.__gt) > 0 else 0
        pq = (
            len(true_positive_pairs) / len(candidate_ids)
            if len(candidate_ids) > 0
            else 0
        )
        candidate_ratio = len(candidate_ids) / self.__size if self.__size > 0 else 0
        return BlockingMetrics(pc, pq, 1 - candidate_ratio)



from abc import abstractmethod
from typing import Protocol

from matchescu.reference_store.comparison_space import BinaryComparisonSpace


class ComparisonSpaceReader(Protocol):
    @abstractmethod
    def read(self) -> BinaryComparisonSpace:
        raise NotImplementedError


class ComparisonSpaceWriter(Protocol):
    @abstractmethod
    def write(self, comparison_space: BinaryComparisonSpace) -> None:
        raise NotImplementedError



from random import shuffle
from typing import Generator, Iterable

from matchescu.comparison_space.generation.blocking import Blocker, Block
from matchescu.reference_store.id_table import IdTable
from matchescu.typing import EntityReferenceIdentifier as RefId


class GroundTruthBlocker(Blocker):
    def __init__(
        self,
        id_table: IdTable,
        ground_truth: Iterable[tuple[RefId, RefId]],
        neg_pos_ratio: float = 8.0,
        max_count: int | None = None,
    ) -> None:
        super().__init__(id_table)
        self._ground_truth = set(ground_truth)
        self._neg_pos_ratio = neg_pos_ratio
        self._max_count = (
            max_count
            if max_count is not None
            else int(len(self._ground_truth) * (1 + self._neg_pos_ratio))
        )

    def _is_true_match(self, x: RefId, y: RefId) -> bool:
        return (x, y) in self._ground_truth or (y, x) in self._ground_truth

    def _get_by_source(self, src: str) -> Generator[RefId, None, None]:
        source_refs = list(self._id_table.get_by_source(src))
        shuffle(source_refs)
        yield from (x.id for x in source_refs)

    def __call__(self) -> Generator[Block, None, None]:
        total_pairs = 0
        per_source_refs = {}
        max_pairs = self._max_count

        def _add_negs(
            blk: Block, x: RefId, compl: RefId, budget: int, n_other: int
        ) -> tuple[int, int]:
            nonlocal total_pairs
            nonlocal per_source_refs

            ret_refs, ret_pairs = 0, n_other
            # lazily cache ref IDs per source
            per_source_refs[x.source] = per_source_refs.get(
                x.source, list(self._get_by_source(x.source))
            )
            for rid in per_source_refs[x.source]:
                if total_pairs >= max_pairs:
                    break
                if rid == x or self._is_true_match(rid, compl):
                    continue
                blk.append(rid)
                ret_refs += 1
                ret_pairs += n_other
                total_pairs += n_other
                if ret_pairs >= budget:
                    break
            return ret_refs, ret_pairs

        for a, b in self._ground_truth:
            if total_pairs >= max_pairs:
                break

            remaining_slots = max_pairs - total_pairs
            if remaining_slots < 1:
                break

            max_negatives = min(int(self._neg_pos_ratio), remaining_slots - 1)
            if max_negatives < 0:
                continue

            # ref IDs from a.source are required in both paths - lazily cache it
            per_source_refs[a.source] = per_source_refs.get(
                a.source, list(self._get_by_source(a.source))
            )
            block = Block(key=(a, b))
            block.append(a).append(b)
            total_pairs += 1

            if a.source != b.source:
                a_refs, a_pairs = _add_negs(block, a, b, max_negatives // 2, 1)
                _add_negs(block, b, a, max_negatives - a_pairs, 1 + a_refs)
            else:
                # a and b are in the current block
                current_size = 2
                added_pairs = 0
                for ref_id in per_source_refs[a.source]:
                    if added_pairs > max_negatives or total_pairs >= max_pairs:
                        break
                    if (
                        ref_id in (a, b)
                        or self._is_true_match(ref_id, a)
                        or self._is_true_match(ref_id, b)
                    ):
                        continue

                    new_pairs = current_size
                    if added_pairs + new_pairs > max_negatives:
                        break

                    block.append(ref_id)
                    added_pairs += new_pairs
                    total_pairs += new_pairs
                    current_size += 1

            yield block



from typing import Generator

from matchescu.reference_store.id_table import IdTable
from matchescu.typing import EntityReference
from ._block import Block
from ._blocker import Blocker
from ._tokenization import tokenize_reference


class SortedNeighborhoodBlocker(Blocker):
    def __init__(self, id_table: IdTable, window_size: int = 4):
        super().__init__(id_table)
        self.__window_size = window_size

    def __sorted_references(self) -> list[tuple[str, EntityReference]]:
        tokenized_refs = list(
            map(lambda r: (" ".join(tokenize_reference(r)), r), self._id_table)
        )
        tokenized_refs.sort(key=lambda t: t[0])
        return tokenized_refs

    def __call__(self) -> Generator[Block, None, None]:
        sorted_refs = self.__sorted_references()
        for i in range(len(sorted_refs) - self.__window_size + 1):
            yield Block(i).extend(
                x[1].id for x in sorted_refs[i : i + self.__window_size]
            )



from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path

import polars as pl
from polars.exceptions import NoDataError

from matchescu.typing import EntityReferenceIdentifier as RefId
from matchescu.reference_store.comparison_space import (
    BinaryComparisonSpace,
    InMemoryComparisonSpace,
)

from ._base import ComparisonSpaceReader, ComparisonSpaceWriter


@dataclass(frozen=True)
class CsvComparisonSpaceFileParams:
    has_header: bool = field(default=False)
    source_fallback: str = field(default="default")
    left_id_column: str | int = field(default=0)
    left_source_column: str | int | None = field(default=None)
    right_id_column: str | int = field(default=1)
    right_source_column: str | int | None = field(default=None)


class CsvPersistence(ComparisonSpaceReader, ComparisonSpaceWriter):
    def __init__(self, file: str | PathLike) -> None:
        self._path = Path(file)

    def read(
        self, params: CsvComparisonSpaceFileParams | None = None
    ) -> BinaryComparisonSpace:
        params = params or CsvComparisonSpaceFileParams()
        cs = InMemoryComparisonSpace()
        try:
            df = pl.read_csv(
                self._path, has_header=params.has_header, ignore_errors=True
            )
        except NoDataError:
            return cs
        col_map = {col: idx for idx, col in enumerate(df.columns)}
        left_id_idx = col_map.get(params.left_id_column, params.left_id_column)
        left_source_idx = col_map.get(
            params.left_source_column, params.left_source_column
        )
        right_id_idx = col_map.get(params.right_id_column, params.right_id_column)
        right_source_idx = col_map.get(
            params.right_source_column, params.right_source_column
        )
        for row in df.iter_rows(named=False):
            left_id = RefId(
                label=row[left_id_idx],
                source=(
                    row[left_source_idx] if left_source_idx else params.source_fallback
                ),
            )
            right_id = RefId(
                label=row[right_id_idx],
                source=(
                    row[right_source_idx]
                    if right_source_idx
                    else params.source_fallback
                ),
            )
            cs.put(left_id, right_id)
        return cs

    def write(self, comparison_space: BinaryComparisonSpace) -> None:
        df = pl.DataFrame(
            [
                {
                    "left_id": left.label,
                    "left_source": left.source,
                    "right_id": right.label,
                    "right_source": right.source,
                }
                for left, right in comparison_space
            ]
        )
        df.write_csv(self._path, include_header=True)



from ._base import ComparisonSpaceReader, ComparisonSpaceWriter
from ._csv import CsvComparisonSpaceFileParams, CsvPersistence


__all__ = [
    "ComparisonSpaceReader",
    "ComparisonSpaceWriter",
    "CsvPersistence",
    "CsvComparisonSpaceFileParams",
]



from typing import Generator

from sklearn.feature_extraction.text import TfidfVectorizer

from matchescu.reference_store.id_table._protocol import IdTable

from ._block import Block
from ._blocker import Blocker
from ._tokenization import tokenize_reference


class TfIdfBlocker(Blocker):
    def __init__(self, id_table: IdTable, min_score: float = 0.1):
        super().__init__(id_table)
        self._min_score = min_score

    def __call__(self) -> Generator[Block, None, None]:
        refs = list(self._id_table)
        corpus = [" ".join(tokenize_reference(ref)) for ref in refs]
        vectorizer = TfidfVectorizer()
        tfidf_matrix = vectorizer.fit_transform(corpus)
        token_inverted_map = vectorizer.get_feature_names_out()
        blocks: dict[str, Block] = {}
        for idx, ref in enumerate(refs):
            tfidf_scores = tfidf_matrix[idx].toarray().flatten()
            for score_idx, score in enumerate(tfidf_scores):
                if score < self._min_score:
                    continue
                score_token = token_inverted_map[score_idx]
                block = blocks.get(score_token, Block(score_token))
                block.append(ref.id)
                blocks[score_token] = block
        yield from blocks.values()
